
# Copyright 2018, 2019 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

import socket
import select
import time
import json
import threading
from netq_lib.common.enums import ConfKw
from netq_lib.orm import HOSTNAME
from netq_lib.common.netq_pq import QueueElem

import grpc
from netq_agent.interface.netq_opta_stream_pb2 import (
    NetqAgentMessage,
)

from netq_agent.interface.netq_opta_stream_pb2_grpc import (
    NetqOPTAStreamServiceStub
)

class AGENTSTATUS():
    SUCCESS = 0
    FAILURE = 1


class ChannelState():
    IDLE = 1
    CONNECTING = 2
    CONNECTED = 3
    DISCONNECTED = 4
    SHUTDOWN = 5

class TransMode():
    PARTIAL_TRANSFORM = 1
    FULL_TRANSFORM = 0
    FIRST_SEGMENT = 2
    MID_SEGMENT = 3
    LAST_SEGMENT = 4

import grpc



RPC_CONNECTIVITY_MAPPING = {}
RPC_CONNECTIVITY_MAPPING[grpc.ChannelConnectivity.IDLE] = ChannelState.IDLE
RPC_CONNECTIVITY_MAPPING[grpc.ChannelConnectivity.CONNECTING] = ChannelState.DISCONNECTED
RPC_CONNECTIVITY_MAPPING[grpc.ChannelConnectivity.READY] = ChannelState.CONNECTED
RPC_CONNECTIVITY_MAPPING[grpc.ChannelConnectivity.TRANSIENT_FAILURE] = ChannelState.DISCONNECTED
RPC_CONNECTIVITY_MAPPING[grpc.ChannelConnectivity.SHUTDOWN] = ChannelState.IDLE


class NetQAgentGrpcClient(object):

    def __init__(self, conf, logger, pq, backpressure_queue):
        self._conf = conf
        self._channel = None
        self._logger = logger
        self.channel_state = ChannelState.IDLE
        self._pq = pq
        self._bp_pq = backpressure_queue
        self.seq_no = 0
        self._max_opta_retry = 3

        try:
            self._num_msg_per_packet = self._conf['netq-agent']['num_msgs_per_packet']
        except:
            self._num_msg_per_packet = 30

        self._logger.info("Number of messagees per packet {}".format(self._num_msg_per_packet))


    def get_sequence_number(self):
        self.seq_no = self.seq_no + 1
        return self.seq_no


    def callback_on_exception(self, exception, traceback):

        self._logger.error(traceback)
        self._logger.error(exception)


    def _run():
        pass

    @property
    def channel_state(self):
        return self._channel_state


    @channel_state.setter
    def channel_state(self, new_state):
        if new_state in RPC_CONNECTIVITY_MAPPING:
            new_state = RPC_CONNECTIVITY_MAPPING[new_state]
        self._logger.debug("New RPC Channel State {}".format(new_state))
        self._channel_state = new_state


    def _get_agent_connect_id(self, verbose=False):
        id = None
        try:
            opta_url = self._conf['netq-agent'][ConfKw.SERVER]
            opta_port = self._conf['netq-agent'].get(ConfKw.PORT, ConfKw.DEFAULT_OPTA_PORT)
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect((opta_url, opta_port))
            id = s.getsockname()[0]
        except Exception:
            self._logger.error("Failed to connect to OPTA")

        if verbose:
            if id:
                self._logger.info("OPTA Discovery Agent Connect IP {}".format(id))
            else:
                self._logger.info("OPTA Discovery No Connect IP determined")
        return id


    def get_agent_id(self):
        if not self.is_valid_hostname(HOSTNAME):
            return None
        return HOSTNAME


    def is_valid_hostname(self, hostname):
        if hostname.lower() in ["cumulus"]:
            return False
        return True


    def full_state_message_builder(self, msg_type, segment, seq_no, json_out):

        netq_message = NetqAgentMessage(aid=socket.gethostname(),
                                        message_type=msg_type,
                                        seqno=seq_no,
                                        ts=long(time.time() * 1000),
                                        trans_mode=segment,
                                        type=0,
                                        message=json.dumps(json_out).encode('utf-8'))
        return netq_message


    def message_builder(self, msg_type, fullstate, json_out):

        full_state = TransMode.FULL_TRANSFORM if fullstate else TransMode.PARTIAL_TRANSFORM

        netq_message = NetqAgentMessage(aid=socket.gethostname(),
                                        message_type=msg_type,
                                        seqno=self.get_sequence_number(),
                                        ts=long(time.time() * 1000),
                                        trans_mode=full_state,
                                        type=0,
                                        message=json.dumps(json_out).encode('utf-8'))
        return netq_message


    def get_full_state_message(self, msg):

        num_msg_per_packet = self._num_msg_per_packet

        try :
            pq_idx = msg['priority_qid']
            msg_type = msg['message_type']
        except Exception as e:
            self._logger.error("Invalid full state message")

        queue_len = self._pq.len(pq_idx)
        num_segments = int(((queue_len-1)/num_msg_per_packet) + 1)

        msg_data = []
        num_msgs = 0

        seq_no = self.get_sequence_number()
        segment_idx = 1

        for priority, message_type, fullstate, msg in self._pq.get_buffer(pq_idx):
            msg_data.append(msg.data)
            num_msgs += 1
            if (num_msgs == num_msg_per_packet):
                if (segment_idx == num_segments):
                    rpc_msg = self.full_state_message_builder(message_type, TransMode.LAST_SEGMENT, seq_no, msg_data)
                elif(segment_idx == 1):
                    rpc_msg = self.full_state_message_builder(message_type, TransMode.FIRST_SEGMENT, seq_no, msg_data)
                else:
                    rpc_msg = self.full_state_message_builder(message_type, TransMode.MID_SEGMENT, seq_no, msg_data)
                msg_data = []
                num_msgs = 0
                segment_idx += 1
                yield rpc_msg

        if len(msg_data) > 0:
            rpc_msg = self.full_state_message_builder(message_type, TransMode.LAST_SEGMENT, seq_no, msg_data)
            yield  rpc_msg


    def get_stream(self):

        pq = self._pq
        fd_set = pq.get_reader_fds()
        log_once = False
        my_run_id = 1

        try:
            (epoll_input, [], []) = select.select(fd_set, [], [], 10)
            if not len(epoll_input):
                return
            if self.get_sequence_number() % 1000 == 0:
                self._logger.info("Current Sequence Number {}".format(self.get_sequence_number()))
            for queue_idx in range(0, self._pq.max_num_queue):
                for priority, message_type, fullstate ,msg in pq.get_buffer(queue_idx):
                    if (message_type == 'FULL_STATE_END'):
                        full_gen = self.get_full_state_message(msg.data)
                        for segment in full_gen:
                            yield  segment
                    if isinstance(msg.data, list):
                        rpc_msg = self.message_builder(message_type, fullstate, msg.data)
                        yield  rpc_msg
                    else:
                        rpc_msg = self.message_builder(message_type, fullstate, [msg.data])
                        yield rpc_msg

        except Exception as e:
            self._logger.error("GRPC %d StreamMessage failed %s", my_run_id, e)
            self.run_id = 0
            self.running_rpc_cnt = 0
            return


    def _get_opta_addr(self):

        serv_addr = "{}:{}".format(self._conf['netq-agent'][ConfKw.SERVER],
                                   self._conf['netq-agent'].get(ConfKw.PORT, ConfKw.DEFAULT_OPTA_PORT))
        return serv_addr


    def monitor_channel(self, state):
        self._logger.debug("Channel State {}".format(state))
        self.channel_state = state


    def reestablish_channel(self):

        serv_addr = self._get_opta_addr()
        self._logger.debug("Reestablishing OPTA Connection {}".format(serv_addr))

        try:
            self._channel = grpc.insecure_channel(serv_addr)
        except Exception as e:
            self._logger.error(
                "Unable to establish connection with OPTA {}".format(e))
            return AGENTSTATUS.FAILURE

        if self._channel:
            self._channel.subscribe(self.monitor_channel, try_to_connect=True)
        else:
            self._logger.error("Unable to connect to OPTA")
            return AGENTSTATUS.FAILURE

        if self.channel_state != ChannelState.CONNECTED:
            return AGENTSTATUS.FAILURE

        return AGENTSTATUS.SUCCESS


    def run_contiguous_stream(self):

        opta_stub = NetqOPTAStreamServiceStub(self._channel)

        while True:
            try:
                stream_response = opta_stub.StreamMessage(self.get_stream())
            except Exception as e:
                self._logger.error("Agent Stream Failed {}".format(e))
                return AGENTSTATUS.FAILURE

    def send_pause_state(self, state):
        try:
            q_elem = QueueElem(0, 'bp', state, 'bp')
            self._bp_pq.putq(q_elem)
            if(state == False):
                for queue_idx in range(0, self._pq.max_num_queue):
                    self._pq.empty_pq(queue_idx)
            self._logger.debug("Agent Connectivity State {}".format(state))
        except Exception as ex:
            self._logger.info('GRPC backpressure putq failed %s', ex)


    def start_stream(self, verbose=False):

        retryCnt = 1

        while True:

            self._logger.info("Agent Stream started")

            if AGENTSTATUS.SUCCESS != self.reestablish_channel():
                self._logger.debug("Failed to establish Connection with OPTA, retrying with backoff")
                if (retryCnt >= self._max_opta_retry):
                    self.send_pause_state(False)
                    retryCnt = 0
                else:
                    retryCnt = retryCnt + 1
                time.sleep(5)
                continue

            self._logger.debug("Starting Stream")
            self.send_pause_state(True)

            try:
                t1 = threading.Thread(target=self.run_contiguous_stream(), args=())
                #stream_response = opta_stub.StreamMessage(self.get_stream())
                t1.start()
                t1.join()
                self._logger.debug("Thread procesisng complete")
            except Exception as e:
                self._logger.error("OPTA Stream Unknown GRPC Exception {}".format(e))
                return AGENTSTATUS.FAILURE

            self._logger.debug("Stream Completed")
